#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2004 Free Software Foundation
#
# FILE:
# GFForm.py
#
# DESCRIPTION:
"""
Class that contains the internal python object representation of a GNUe Form
built from GFObjects
"""
# NOTES:
#

import sys
import string
import traceback

from gnue.common.apps import GDebug
from gnue.common import events
from gnue.common.datasources.GDataObjects import ConnectionError as DBError
from gnue.common.definitions.GRootObj import GRootObj
from gnue.common.logic.language import AbortRequest
from gnue.common.definitions.GObjects import GObj
from gnue.common.datasources import ConnectionTriggerObj
from gnue.forms.GFObjects import *
from gnue.forms import GFParser


# Defines which objects are "Tab Stops"
TabStops = ('GFEntry','GFButton')

#
# GFForm
#
# instance is passed into the initializer so that
# designer can pass that in
#
class GFForm(GRootObj, GFObj, events.EventAware):
  def __init__(self, parent=None, instance=None):
  # TODO: with a little tweaking we can now stop passing  GFParser.getXMLelements
    GRootObj.__init__(self, 'form', GFParser.getXMLelements, GFParser)
    GFObj.__init__(self, parent)

    self._type = "GFForm"
    self.name="__main__"

    # Datasources referced by this form
    self._datasourceDictionary = {}

    # Dictionary to locate named triggers
    self._triggerDictionary = {}

    # Insert/Overwrite mode
    self._insertMode = True

    # Focus information
    self._currentPage = None
    self._currentBlock = None
    self._currentEntry = None

    # set
    self._instance = instance
    self._app = self._instance #TODO: Imports are broken do to
                               #TODO: switch from _app to _instance

    # Hackery until proper layout support is added
    self._standardnamespaces = {'Char': 'GNUe:Forms:Char'}


    # The "None" init gives datasources time to setup master/detail
    self._inits = [self.primaryInit, None, self.secondaryInit]

    #
    # Trigger support
    #
    self._triggerns = {}
    self._triggerGlobal = True # Put this object into the global namespace by it's name

    self._validTriggers = { 'ON-STARTUP':     'On-Startup',
                            'ON-ACTIVATION':  'On-Activation',
                            'PRE-EXIT':       'Pre-Exit',
                            'ON-EXIT':        'On-Exit',
                            'PRE-COMMIT':     'Pre-Commit',
                            'POST-COMMIT':    'Post-Commit' }

    self._triggerFunctions = {'setFocus':{'function':self.triggerSetFocus,
                                          'global': True,
                                          },
                              'getAuthenticatedUser':{'function':self.triggerGetAuthenticatedUser,
                                                      'global': True,
                                                      },
                              'setTitle':{'function':self.triggerSetTitle,
                                                      'global': True,
                                                      },
#                               'getCurrentEntry':{'function':self.triggerGetCurrentEntry,
#                                                  'global': True,
#                                                  },
#                               'getCurrentBlock':{'function':self.triggerGetCurrentBlock,
#                                                  'global': True,
#                                                  },
#                               'getCurrentPage':{'function':self.triggerGetCurrentPage,
#                                                 'global': True,
#                                                },
                              'setStatusText':{'function':self.triggerSetStatusText,
                                               'global': True,
                                               },
                              'getParameter':{'function':self.getParameter,
                                              'global': True,
                                              },
                              'setParameter':{'function':self.triggerSetParameter,
                                              'global': True,
                                              },
                              'showMessage':{'function':self.triggerShowMessageBox,
                                             'global': True,
                                             },
                              'messageBox': {
                                   'function': self.triggerMessageBox,
                                   'global': True},
                              'commit':{'function':self.commit,
                                        'global': True,
                                        },
                              'rollback':{'function':self.rollback,
                                        'global': True,
                                        },
                              'close':{'function':self.triggerClose,
                                       'global': True,
                                       },
                              'getFeature':{'function':self.triggerGetFeature,
                                            'global': True,
                                            },
                              'setFeature':{'function':self.triggerSetFeature,
                                            'global': True,
                                            },
                              'runForm':{'function':self.triggerRunForm,
                                         'global': True,
                                         } ,
                              'activateDialog':{'function':self.triggerActivateDialog,
                                                'global': True,
                                                } ,
                              'activateTrigger':{'function':self.fireTrigger,
                                                'global': True,
                                                } ,
                              'initQuery':{'function':self.initQuery,
                                           },
                              }

    self._features = {
      'GUI:MENUBAR:SUPPRESS': False,
      'GUI:TOOLBAR:SUPPRESS': False,
      'GUI:STATUSBAR:SUPPRESS': False,
    }

    self._in_trigger_lock = False

  #
  # primaryInit
  #
  # Called during phaseInit startup by GParser.
  #
  def primaryInit(self):
    # Initialize our events system
    events.EventAware.__init__(self, self._instance.eventController)

    # Find the logic and layout controllers
    for child in self._children:
      if child._type == 'GFLogic':
        self._logic = child
      elif child._type == 'GFLayout':
        self._layout = child

    self.initTriggerSystem()
    ConnectionTriggerObj.addAllConnections(self._instance.connections, self._triggerNamespaceTree)
    self._triggerns.update(self._triggerNamespaceTree._globalNamespace)

    ## TODO: This ain't right!  Fix after 0.5.0
    #self._triggerns['runForm'] = self.triggerRunForm
    ## TODO


  #
  # Secondary init (stuff that must happen
  # after all our children have init'ed)
  #
  def secondaryInit(self):
    # create the first records
    for key in self._datasourceDictionary.keys():
      if not self._datasourceDictionary[key].hasMaster():
        self._datasourceDictionary[key].createEmptyResultSet()

    # Set initial focus
    self.findAndChangeFocus(self)
    self.processTrigger('On-Startup')

  #
  # Get a user parameter. If parameter not specified, pull default value
  #
  def getParameter(self, parameter):
    param = string.lower(parameter)
    try:
      return self._parameters[param]
    except KeyError:
      rv = None
      for child in self._children:
        if isinstance(child, GFParameter) and child._name == param:
          rv = child.default
          self._parameters[param] = rv
          break
    return rv

  ###################################################################
  #
  # Focus functions
  #
  #

  #
  # findAndChangeFocus
  #
  # Runs findFocus and then runs
  # changeFocus with that value.
  #
  def findAndChangeFocus(self, object):
    entry = self.findFocus(object)
    if entry:
      self.changeFocus(entry)

  #
  # Find focus
  #
  # Find the next focusable item given
  # an item of type Page, Block, or Entry
  #
  # Only called by findAndChangeFocus
  #
  def findFocus(self, object):

    if object._type == 'GFField':
      GDebug.printMesg(0,"GFField was passed to 'findFocus'.")
      try:
        object = object._entryList[0]
      except IndexError:
        return None

    entry = None

    mode = self.getCurrentMode()

    if isinstance(object, GFObj):
      if object.isNavigable(mode):
        return object
      else:
        if hasattr(object, 'getFocusOrder'):
          for child in object.getFocusOrder():
            entry = self.findFocus(child)
            if entry:
              return entry
        #
        # We're in trouble.  Scan all children for a
        # focusable widget.  Will dump GFField objects
        # stored in blocks into itself.
        #
        for child in object._children:
          entry = self.findFocus(child)
          if entry:
            break

    return entry

  #
  #
  #
  def getCurrentMode(self):
    """
    Returns the editing mode of the current record
    """
    if self._currentBlock:
      if self._currentBlock.mode == 'query':
        return 'query'
      elif self._currentBlock.isEmpty():
        return 'new'
      else:
        return 'edit'
    else:
      return 'new'

  #
  # changeFocus
  #
  # changes to the requested entry object requested by
  # an event source
  #
  def changeFocus(self, widget, fireFocusTriggers=1):

    message = None

    # cannot change focus to same entry
    if widget == self._currentEntry:
      GDebug.printMesg (1, "Widget is current entry %s" % widget)
      return

    cen = self._currentEntry is not None and self._currentEntry.name or '-'
    win = widget is not None and widget.name or '-'
    GDebug.printMesg (1, "Changing focus from %s (%s) to %s (%s)" % \
        (self._currentEntry, cen, widget, win))

    try:
      if self._currentEntry:

        if self._currentEntry._type != 'GFButton':
          val = self._currentEntry.getValue()
          if self._currentEntry._field.minLength and val is not None and len(str(val)):
            if len(str(val)) < self._currentEntry._field.minLength:
              message = u_("Minimum required length %d") \
                        % self._currentEntry._field.minLength
              # Piggybacking off the triggers message box code
              self.triggerShowMessageBox (message)
              return

        event = events.Event('endEDITMODE',_form=self)
        self.dispatchEvent(event)

        #self._instance.dispatchEvent(event)
        if event.__error__:
          return True

      fieldChange = widget != self._currentEntry
      try:
        blockChange = widget._block != self._currentBlock
      except AttributeError:
        # Buttons don't have a block, but also
        # don't trigger a block change
        blockChange = False
      pageChange = widget._page != self._currentPage

      if fireFocusTriggers:
        try:
          if fieldChange:
            self._currentEntry.processTrigger('Pre-FocusOut', ignoreAbort=False)
            if hasattr(self._currentEntry,'_field'):
              self._currentEntry._field.processTrigger('Pre-FocusOut', ignoreAbort=False)
          if blockChange:
            self._currentBlock.processTrigger('Pre-FocusOut', ignoreAbort=False)
          if pageChange:
            self._currentPage.processTrigger('Pre-FocusOut', ignoreAbort=False)
            self._currentPage.processTrigger('Post-FocusOut', ignoreAbort=False)
          if blockChange:
            self._currentBlock.processTrigger('Post-FocusOut', ignoreAbort=False)
          if fieldChange:
            self._currentEntry.processTrigger('Post-FocusOut', ignoreAbort=False)
            if hasattr(self._currentEntry,'_field'):
              self._currentEntry._field.processTrigger('Post-FocusOut', ignoreAbort=False)
        except AttributeError:
          pass

      oldEntry = self._currentEntry

      self._currentEntry = widget
      try:
        self._currentBlock = self._currentEntry._block
      except AttributeError:
        pass # Buttons, et al
      self._currentPage = self._currentEntry._page

      if pageChange:
        self.dispatchEvent('gotoPAGE',self._currentPage, _form=self);

      if blockChange:
        self.refreshDisplay(self._currentBlock)

      GDebug.printMesg (1, "Updating entries old: %s, new: %s" % \
          (oldEntry, self._currentEntry))

      self.dispatchEvent('updateENTRY', oldEntry, _form=self)
      self.dispatchEvent('updateENTRY', self._currentEntry, _form=self)

      if fireFocusTriggers:
        if pageChange:
          self._currentPage.processTrigger('Pre-FocusIn', ignoreAbort=False)
        if blockChange:
          self._currentBlock.processTrigger('Pre-FocusIn', ignoreAbort=False)
        if fieldChange:
          self._currentEntry.processTrigger('Pre-FocusIn', ignoreAbort=False)
          if hasattr(self._currentEntry,'_field'):
            self._currentEntry._field.processTrigger('Pre-FocusIn', ignoreAbort=False)
          if hasattr(self._currentEntry,'_field'):
            self._currentEntry._field.processTrigger('Post-FocusIn', ignoreAbort=False)
          self._currentEntry.processTrigger('Post-FocusIn', ignoreAbort=False)
        if blockChange:
          self._currentBlock.processTrigger('Post-FocusIn', ignoreAbort=False)
        if pageChange:
          self._currentPage.processTrigger('Post-FocusIn', ignoreAbort=False)

      GDebug.printMesg (1, "calling refreshUIEvents ()")
      self.refreshUIEvents()
      GDebug.printMesg (1, "CurrentEntry is %s" % self._currentEntry)

    except AbortRequest, t:
      GDebug.printMesg(1, "Trigger Error!")
      message = _("Trigger Exception :\n") + t.message
    return message


  #
  #
  # Events
  #
  #

  #
  # Incoming Event handlers
  #

  def newRecord(self):
    if not self.readonly:
      self._currentBlock.newRecord()

  def deleteRecord(self):
    if not self.readonly:
      self._currentBlock.deleteRecord()
      self.refreshUIEvents()

  def isSaved(self):
    if (self._currentEntry._type != 'GFButton' and \
        self._currentEntry._displayHandler.isPending()):
      return False

    for block in self._logic._blockList:
      if not block.isSaved():
        return False

    return True

  def initQuery(self):
    message = None
    if self._currentBlock.mode != 'query':

      # Auto-commit?
      if self._currentBlock.autoCommit:
        self.commit()

      for block in self._logic._blockList:
        if not block.isSaved():
          message = _('Data not saved. Save changes or clear the form to proceed.')
          return message

    for block in self._logic._blockList:
      block.processRollback()
      for entry in block._entryList:
        if hasattr(entry,'queryDefault'):
          GDebug.printMesg(1, "%s will be set to %s" % (entry.name, entry.queryDefault))
          entry.setValue(entry.queryDefault)

    # If Enter-Query is hit once, enter query mode
    # If Enter-Query is hit twice, bring back conditions from last query.
    # If Enter-Query is hit thrice, cancel the query and go into normal mode.

    self.dispatchEvent('beginENTERQUERY', _form=self)
    #self._instance.dispatchEvent('beginENTERQUERY', _form=self)

    for block in self._logic._blockList:
      block.initQuery()

    self.triggerSetStatusText(_('Enter your query criteria.'))
    return message


  def cancelQuery(self):
    self.dispatchEvent('endENTERQUERY', _form=self)
    #self._instance.dispatchEvent('endENTERQUERY', _form=self)
    message = None
    for block in self._logic._blockList:
      block.cancelQuery()

    self.triggerSetStatusText(_('Query canceled.'))
    return message

  def copyQuery(self):
    if self._currentBlock.mode != 'query':
      self.initQuery()

    self.dispatchEvent('endENTERQUERY', _form=self);
    #self._instance.dispatchEvent('endENTERQUERY', _form=self);
    message = None
    for block in self._logic._blockList:
      block.copyQuery()

    return message


  def executeQuery(self):
    if self._currentBlock.mode != 'query':
      return _("Form not in query mode")

    self.dispatchEvent('endENTERQUERY', _form=self);
    self.dispatchEvent('beginWAIT',None, _form=self);
    #self._instance.dispatchEvent('endENTERQUERY', _form=self);
    #self._instance.dispatchEvent('beginWAIT',None, _form=self);

    message = None
    try:
      self._currentBlock.processQuery()

    except DBError:
      self.rollback(1)
      message = _("Database query error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])


    self.dispatchEvent('endWAIT', _form=self)
    #self._instance.dispatchEvent('endWAIT', _form=self)
    # TODO: may need to make self._currentBlock just self
    self.refreshDisplay(self._currentBlock)
    if self._currentBlock.isEmpty():
      self.triggerSetStatusText(_('Query returned no results.'))
    else:
      self.triggerSetStatusText(_('Query sucessful.'))




  def commit(self):
    self.dispatchEvent('beginWAIT', _form=self)
    #self._instance.dispatchEvent('beginWAIT', _form=self)

    message = None

    try:
      if not self.readonly:
        # Form level pre-commit
        try:
          self.processTrigger('Pre-Commit', ignoreAbort=False)
        except AbortRequest:
          GDebug.printMesg(1, "Trigger form Pre-Commit threw a AbortRequest!")
          self.dispatchEvent('endWAIT',None, _form=self)
          #self._instance.dispatchEvent('endWAIT',None, _form=self)
          return _("Form trigger returned error")

        for block in self._logic._blockList:
          GDebug.printMesg(1, "Saving %s"%block.name)
          try:

            # This gets lost in the Pre-Commit code
            block._precommitRecord = block._currentRecord

            block.processCommit()
          except AbortRequest:
            GDebug.printMesg(1, "Trigger block Pre-Commit threw a AbortRequest!")
            self.dispatchEvent('endWAIT',None, _form=self)
            #self._instance.dispatchEvent('endWAIT',None, _form=self)
            return _("Block trigger returned error")
          block.processTrigger('Post-Commit')

        for block in self._logic._blockList:
          if block.autoClear:
            block.processClear()
          else:
            GDebug.printMesg (1, "switchRecord (0) from commit")
            block.switchRecord(0)

        self.dispatchEvent('cannotCOMMIT')
        self.dispatchEvent('cannotROLLBACK')

      else:
        message = _('Form is readonly')
        #self._instance.dispatchEvent('endWAIT',None, _form=self)
        self.dispatchEvent('endWAIT',None, _form=self)

        return message

      self.processTrigger('Post-Commit')

    except DBError:
      self.rollback(1)
      message = _("Database commit error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])
    except:
      print _("\n\nGFForm: Unexpected Exception:")
      print '-'*60
      traceback.print_exc(file=sys.stdout)
      print '-'*60

    #self._instance.dispatchEvent('endWAIT',None, _form=self)
    self.dispatchEvent('endWAIT',None, _form=self)

    return message

  def rollback(self, recover=0):
    self.endEditing()
    for block in self._logic._blockList:
      block.processRollback(recover)
    self.refreshDisplay(self)
    self._currentBlock.jumpRecord(self._currentBlock._currentRecord)
    self.dispatchEvent('cannotCOMMIT')
    self.dispatchEvent('cannotROLLBACK')

  #
  # fireTrigger
  #
  def fireTrigger(self, triggerName):
    self._triggerDictionary[triggerName](self)


  #
  #
  #
  def updateUIEntry(self,field):
    for entry in field._entryList:
      self.dispatchEvent('updateENTRY',entry, _form=self)
      #self._instance.dispatchEvent('updateENTRY',entry, _form=self)


  #
  # nextEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def nextEntry(self, reverse=False, first=False, onlyInBlock=False):

    currentBlock = self._currentBlock
    mode = self.getCurrentMode()


    if currentBlock.transparent and not ( \
          onlyInBlock or \
          currentBlock.autoNextRecord and not ( \
              currentBlock.isEmpty() or \
              (not reverse and currentBlock.isLastRecord() and \
                   not (currentBlock.autoCreate and \
                        currentBlock.editable in ('Y','new')) or \
              (reverse and currentBlock.isFirstRecord()) \
          ))):
      source = self._currentEntry._page.getFocusOrder()
      stayInBlock = False
    else:
      source = currentBlock.getFocusOrder()
      stayInBlock = True

    # If we want the previous entry, then reverse the focusorder we're using
    if reverse:
      source.reverse()

    nextEntry = None
    firstEntry = None
    keepNext = False

    for object in source:

      if object.isNavigable(mode):
        if stayInBlock and \
           (currentBlock.name != object.block):
          continue
        # If we only wanted the first navigable field in the block, then return
        if first:
          nextEntry = object
          break

        # Put the first field as the next to rollover
        if nextEntry == None:
          nextEntry = object
          firstEntry = object

        # If we're at the current focused entry,
        # then the next entry will be what we want
        if object == self._currentEntry:
          keepNext = True

        # If we've already passed the current entry
        # Then this is the entry to return
        elif keepNext:
          nextEntry = object
          break

    # If we've cycled back around to the first entry, then do special checks
    if nextEntry == firstEntry:

      # If we should navigate to the next record, do it...
      if reverse and not currentBlock.isFirstRecord():
        currentBlock.prevRecord()
        self.changeFocus(nextEntry)
      elif not reverse and \
         currentBlock.autoNextRecord and \
         not currentBlock.isEmpty() and \
         not (not currentBlock.autoCreate and \
              currentBlock.isLastRecord()):
           currentBlock.nextRecord()

           # If new record is empty, then the
           # nextEntry logic has been taken care of...
           if currentBlock.isEmpty():
             return
           else:
             self.changeFocus(nextEntry)

      # Otherwise, are we transparent? If so, go to next block/page
      elif currentBlock.transparent and \
         self._currentPage.transparent:

        # Jump to the next/(previous) page if block is page as transparent
        pages =  self._layout._pageList
        i = pages.index(self._currentPage)

        if reverse:
          try:
            dest = self._layout._pageList[i-1]
          except IndexError:
            dest = self._layout._pageList[-1]
          # TODO: this fails if last entry is not navigable
          self.findAndChangeFocus(dest._entryList[-1])
        else:
          try:
            dest = self._layout._pageList[i+1]
          except IndexError:
            dest = self._layout._pageList[0]
          self.findAndChangeFocus(dest)
      else:
        self.changeFocus(nextEntry)

    else:
      self.changeFocus(nextEntry)

  #
  # previousEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def previousEntry(self):
    return self.nextEntry(reverse=True)

  #
  # refreshDisplay
  #
  # Refreshes all the UI fields to match
  # the virtual form's fields
  #
  def refreshDisplay(self, block):
    block.walk(self.__refreshDisplay)
    self.refreshUIEvents()

  def __refreshDisplay(self, object):
    if object._type in ('GFEntry', "GFImage"):
      self.dispatchEvent('updateENTRY',object, _form=self)

  #
  # Signal the UI Drivers of navigation button relevance
  #
  def refreshUIEvents(self):
    block = self._currentBlock
    if not block: return

    dispatchEvent = self.dispatchEvent
    if block.mode == 'query':
      dispatchEvent('canCANCELQUERY')
      dispatchEvent('canEXECQUERY')
      dispatchEvent('cannotENTERQUERY')
      dispatchEvent('cannotPREVRECORD')
      dispatchEvent('cannotFIRSTRECORD')
      dispatchEvent('cannotNEXTRECORD')
      dispatchEvent('cannotLASTRECORD')
      dispatchEvent('cannotCOMMIT')
      dispatchEvent('cannotROLLBACK')
      dispatchEvent('cannotJUMPPROMPT')
      dispatchEvent('cannotNEWRECORD')
      dispatchEvent('cannotMARKFORDELETE')
    else:
      dispatchEvent('canENTERQUERY')
      dispatchEvent('cannotCANCELQUERY')
      dispatchEvent('cannotEXECQUERY')
##      dispatchEvent('canNEXTRECORD')

      # TODO: There should probably be some tests here....
      if not self.isSaved():
        dispatchEvent('canCOMMIT')
      else:
        dispatchEvent('cannotCOMMIT')

      dispatchEvent('canROLLBACK')
      dispatchEvent('canJUMPPROMPT')
      dispatchEvent('canNEWRECORD')
      dispatchEvent('canMARKFORDELETE')

      if block._resultSet.isFirstRecord():
        dispatchEvent('cannotPREVRECORD')
        dispatchEvent('cannotFIRSTRECORD')
      else:
        dispatchEvent('canPREVRECORD')
        dispatchEvent('canFIRSTRECORD')

      if block._resultSet.isLastRecord():
        dispatchEvent('cannotLASTRECORD')
        if block._resultSet.current.isEmpty():
          dispatchEvent('cannotNEXTRECORD')
          dispatchEvent('cannotNEWRECORD')
        else:
          dispatchEvent('canNEWRECORD')
      else:
        dispatchEvent('canNEXTRECORD')
        dispatchEvent('canLASTRECORD')
        dispatchEvent('canNEWRECORD')

  #
  # nextBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def nextBlock(self):

    try:
      nextBlock = self._logic._blockList[self._logic._blockList.index(self._currentBlock)+1]
    except IndexError:
      nextBlock = self._logic._blockList[0]
    self.findAndChangeFocus(nextBlock)



  #
  # findPreviousBlock
  #
  # Convenience routine to find the previous block
  #
  def findPreviousBlock(self):
    try:
      return self._logic._blockList[self._logic._blockList.index(self._currentBlock)-1]
    except IndexError:
      return self._logic._blockList[-1]


  #
  # previousBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def previousBlock(self):
    self.findAndChangeFocus(self.findPreviousBlock())


  #
  # endEditing
  #
  # Signal to the current Entry to stop editing
  # mode and save it's value to the virtual form
  #
  def endEditing(self):
    event = events.Event('endEDITMODE',None, _form=self)
    self.dispatchEvent(event)
    return not event.__error__


  def prevRecord(self):
    if self._currentBlock.mode == 'query':
      self.triggerSetStatusText(_('You cannot do that in query mode.'))
      return
    self._currentBlock.prevRecord()

  def firstRecord(self):
    if self._currentBlock.mode == 'query':
      self.triggerSetStatusText(_('You cannot do that in query mode.'))
      return
    self._currentBlock.firstRecord()

  def lastRecord(self):
    if self._currentBlock.mode == 'query':
      self.triggerSetStatusText(_('You cannot do that in query mode.'))
      return
    self._currentBlock.lastRecord()

  def nextRecord(self):
    if self._currentBlock.mode == 'query':
      self.triggerSetStatusText(_('You cannot do that in query mode.'))
      return
    self._currentBlock.nextRecord()

  def jumpRecord(self,count):
    if self._currentBlock.mode == 'query':
      self.triggerSetStatusText(_('You cannot do that in query mode.'))
      return
    self._currentBlock.jumpRecord(count)

  def toggleInsertMode(self):
    self._insertMode = not self._insertMode


  ###################################################################
  #
  # Trigger functions
  #
  # The following functions are only used to implement
  # functions exposed in the trigger namespace.
  #
  #

  #
  # triggerShowMessageBox
  #
  # display a standard message box
  #
  def triggerShowMessageBox (self, msgtxt, caption='GNUe Message',
      title='Information', kind = 'Info', cancel = False):
    return self._instance.displayMessageBox (msgtxt, kind, cancel, caption,
                                             title)


  # ---------------------------------------------------------------------------
  # Show a message box using the given message
  # ---------------------------------------------------------------------------

  def triggerMessageBox (self, message, kind = None, title = None,
      cancel = False):
    """
    This function brings up a message box of a given kind.
    @param message: text to be displayed
    @param kind: 'Question', 'Info', 'Warning', or 'Error'
    @param cancel: Boolean flag indicating wether a cancel button will be
        included or not.
    @return: True for <Yes> or <Ok> button, False for <No> button, None for
        <Cancel> button.
    """
    return self._instance.displayMessageBox (message, kind, cancel, '', title)

  #
  # triggerSetFocus
  #
  # allow the trigger to switch input focus
  # to a specfic entry widget
  #
  def triggerSetFocus(self,object):
    # add global focus locking
    if self._in_trigger_lock:
      print "Already called by a trigger"
      return

    self._in_trigger_lock = True

    focus = object._object
    if focus._type=='GFField':
      try:
        focus = focus._entryList[0]
      except KeyError:
        raise "setFocus failed: GFField is not bound to an GFEntry object."

    if focus._type != 'GFEntry':
      raise "setFocus failed: Can just switch to GFEntry objects. " +\
            "You passed a '%s' object." % focus._type
    self.changeFocus(focus,1)
    self.dispatchEvent('gotoENTRY',object=self._currentEntry,_form=self)

    self._in_trigger_lock = False

#   #
#   # allow the trigger to get a handle to
#   # an actual GFEntry object
#   #
#   # TODO: WHY?  This seems like a good way to write
#   # TODO: triggers that break due to internal forms changes
#   #
#   def triggerGetCurrentEntry(self):
#     return self._currentEntry
#
#
#   #
#   # allow the trigger to get a handle to
#   # an actual GFBlock object
#   #
#   # TODO: WHY?  This seems like a good way to write
#   # TODO: triggers that break due to internal forms changes
#   #
#   def triggerGetCurrentBlock(self):
#     return self._currentBlock
#
#   #
#   # allow the trigger to get a handle to
#   # an actual GFPage object
#   #
#   # TODO: WHY?  This seems like a good way to write
#   # TODO: triggers that break due to internal forms changes
#   #
#   def triggerGetCurrentPage(self):
#     return self._currentPage

  #
  # triggerSetStatusText
  #
  # Allows the trigger to display a custom message
  # on the forms status bar
  #
  def triggerSetStatusText(self,tip=''):
    self._instance.updateStatusBar(tip=tip, form=self)

  #
  # triggerRunForm
  #
  # Allows the trigger to launch a new instance
  # of gnue-forms running a different form.
  #
  def triggerRunForm(self, fileName):
    from GFInstance import GFInstance
    instance = GFInstance(self._instance.manager,
                          self._instance.connections,
                          self._instance._uimodule,  1)
    instance.addFormFromFile(fileName)
    instance.activate()

  #
  # triggerActivateDialog
  #
  # Allows the trigger to launch a standard or custom
  # dialog.  Values can be passed back and forth thru
  # the parameters dictionary
  #
  def triggerActivateDialog(self,dialogName, parameters={}, modal=0):
    self._instance.activateForm(dialogName,parameters, modal)

  #
  # triggerClose
  #
  # Allows the trigger to close this copy of gnue-forms
  #
  def triggerClose(self):
    event = events.Event('requestEXIT', _form=self)
    #self._instance.dispatchEvent(event)
    self.dispatchEvent(event)
    if event.__error__:
      return 1

  #
  # triggerSetFeature
  #
  # Allows the trigger to set feature values
  # Features are things like toolbars, menubars, and statusbar
  #
  def triggerSetFeature(self, feature, value):
    if not self._features.has_key(feature):
      raise KeyError, "Trigger attempted to set unknown feature %s" % feature
    else:
      self._features[feature] = value

  #
  # triggerGetFeature
  #
  # Allows the trigger to get feature values
  # Features are things like toolbars, menubars, and statusbar
  #
  def triggerGetFeature(self, feature):
    try:
      return self._features[feature]
    except KeyError:
      raise KeyError, "Trigger attempted to get unknown feature %s" % feature

  #
  # triggerGetAuthenticatedUser
  #
  # TODO: Description
  #
  def triggerGetAuthenticatedUser(self, connection=None):
    return self._instance.connections.getAuthenticatedUser(connection)

  #
  # triggerSetTitle
  #
  # TODO: Description
  #
  def triggerSetTitle(self, title):
    self.title = title # assignment of title value passed to form
    # code to fire event to update the display goes here
    self.dispatchEvent('setTitle',title=title, _form=self);


  #
  # Set a user parameter.
  #
  def triggerSetParameter(self, parameter, value):
    param = string.lower(parameter)
    self._parameters[param]=value

  def triggerActivateTrigger(self,name):
    self.processTrigger(name)
